/***************************************************************************
                          basemldocument.cpp  -  description
                             -------------------
    begin                : Sat Dec 8 2001
    copyright            : (C) 2001-2002 by Anthony Liekens
    email                : mooby@nerdhero.org
    Modified by          : Walter Schreppers
    email                : wschrep@uia.ac.be

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "basemldocument.h"
#include "../config.h"

#include <stdlib.h>
#include <iostream>
#include <ctime>
#include <string>
#include <fstream>

//todo replace this with a popen...
//#include <procbuf.h>
#include <unistd.h>

Defines BaseMLDocument::defines;


BaseMLDocument::~BaseMLDocument(){
  //clean up our mess !
  if ( (tree==NULL) || (tree==0) ) delete tree;
  if ( (importDoc==NULL) || (importDoc==0) ) delete importDoc;
}


BaseMLDocument::BaseMLDocument( const string& fileName, const string& childName ) {
  importDoc=NULL;
  recursion=0; //static global!
  indent=0;    //static global!
  tree=NULL;
  if( childName != "" )
  	this->fileName = childName;
  else
	  this->fileName = fileName;

// todo !!!!	
//	QRegExp pathCompacter( "/[^/]*/\\.\\." );
//	while( this->fileName.findRev( pathCompacter ) != -1 ) {
//		this->fileName.replace( pathCompacter, "" );
//	}
	
  //extract path from fileName
  string::size_type pos=fileName.rfind("/");
  if(pos!=string::npos){
    filePath=fileName.substr(0, pos+1 );
  }
  else{
    filePath="";
  }


  ifstream in(fileName.c_str()); 

  if ( !in ) {
  		cerr << "Fatal Error: " << fileName << " cannot be opened for reading." << endl;
      exit( 1 );
  }
  
  if ( !xml.parse(&in) ) {
      cerr << "Fatal Error: " << fileName << " cannot be used as XML document." << endl;
      exit( 1 );
  }
  else{
    tree=xml.tree;
  }
  
  // if the main document has an extend="foo.bml" attribute, we should use that as a parent

  parentDocument = NULL;

  if( (tree==0) || (tree==NULL) ) cerr<< "Fatal Error: "<< fileName << "did not generate a dom tree."<<endl;

  ElementNode* rootNode;
  if( tree->nodeType() == XMLNode::ElementNode ){
    rootNode=(ElementNode*) tree; //cast it to correct type
    if( rootNode->tagName() != "document" ) {
      cerr << "Fatal Error: " << fileName << " should contain a root element with \"document\" tag." << endl;
	  	exit( 1 );
    }
  }
  else{
    cerr << "Fatal Error: " << fileName << " should contain a root element with \"document\" tag." << endl;
  	exit( 1 );
  }


  HTMLFilterDisabled = false;
 	if( rootNode->attribute( "html", "" ) == "disabled" )
 		HTMLFilterDisabled = true;
 	
 	PHPFilterDisabled = true;
 	if( rootNode->attribute( "php", "" ) == "enabled" )
 		PHPFilterDisabled = false;
 	
 	// check whether for <import> elements, and import their documents
  
  for(XMLNode::iterator i=tree->begin();i!=tree->end();i++){
    if( (*i)->nodeType() == XMLNode::ElementNode ){
      ElementNode* e=(ElementNode*) *i; //it was an element so cast it
      if( e->tagName() == "import" ) {
        string importName = e->attribute ( "document", "" );
        if( importName == "" ) {
					cerr << "Warning: No \"document\" attribute given for import element" << endl;
				} else {
					importDocument( checkPath( importName ) );
				}
      }
    }//if
  }//for
}




bool
BaseMLDocument::hasParent() {
	if( parentDocument != NULL ) {
		return true;
	} else {
    if( (tree!=0) && ( tree->nodeType() == XMLNode::ElementNode ) ){
      ElementNode* rootNode=(ElementNode*) tree; //cast it to correct type
 		  if( rootNode->attribute( "extend", "" ) != "" ) {
 	      string fname= rootNode->attribute( "extend", "" );
 			  parentDocument = new BaseMLDocument( checkPath( fname ), fileName );
 	      //made recursion and indent static global int's
        //to avoid using slots (anyone have a better solution?)
 			  return true;
 		  } else {
 			  return false;
 		  }
    }
 	}
  return false;
}




void BaseMLDocument::importDocument( const string& importName ){
	ifstream in( importName.c_str() );
  if( !in ) {
 		cerr << "Import Warning: " << importName << " cannot be opened for reading." << endl;
 	  return;
  } else
  if( !xml.parse(&in) ) {
		cerr << "Import Warning: " << importName << " cannot be used as XML document." << endl;
	} else {
		// imported file is opened - copy all its definitions
	  importDoc=xml.tree;

    for(XMLNode::iterator i=importDoc->begin();i!=importDoc->end();i++){
      if((*i)->nodeType()==XMLNode::ElementNode ){  //see if node is an element and convert it
				tree->appendChild( *i );
      }
    }    
    //delete importDoc; //DO NOT DO THIS! WE KEEP THE POINTERS ALIVE FOR API-USER !!!
  }
}



ElementNode*
BaseMLDocument::findSub( const string &subName ) {
  if(tree!=0){
    //tree->showTree(tree);
  }
  else{
    cerr<<"Fatal Error: Document did not contain tree"<<endl;
    exit(1);
  }
  for(XMLNode::iterator i=tree->begin();i!=tree->end();i++){
    if( (*i)->nodeType()==XMLNode::ElementNode ){ //convert n to element
      ElementNode* e=(ElementNode*) *i; //it was an element so cast it
			if( e->tagName() == subName ) {
				ElementNode* ret = e ;  //maybe use copy constructor instead...
				return ret;
			}
		}
  }
  if( hasParent() ) {
  	return parentDocument->findSub( subName );
  }
	return 0;
}



void
BaseMLDocument::renderContainer( ostream& os, ElementNode* sub, ElementNode* parent, Stack& callers, Attributes& attributes, const string& action ) {
  if((sub==0)||(sub==NULL)){
    cerr<<"Error in renderContainer, sub=0!"<<endl;
    return;
  }
  for(XMLNode::iterator i=sub->begin();i!=sub->end();i++){
		if( (*i)->nodeType()==XMLNode::TextNode ) { //if node is a text node
			// we only want <item>s in containers,
			// check whether this text node only contains white space, then it's ok
			TextNode* e=(TextNode*) *i;
			if(e!=0){
				string data = e->data();
			  if( removeWhiteSpace( data ) != "" )
			  	cerr << "Warning: A container may only contain \"item\" elements, all others are ignored" << endl;
			}
		} else if( (*i)->nodeType()==XMLNode::ElementNode ) {
	    ElementNode* e=(ElementNode*) *i; //convert it to an element
			if( e!=0 ) { 
				string tagName = e->tagName();
				if( tagName != "item" ) {
					cerr << "Warning: A container may only contain \"item\" elements, all others are ignored" << endl;
				} else {
					if( e->attribute( "document", "" ) == "" ) {
						cerr << "Warning: Cannot render item without \"document\" attribute" << endl;
					} else {
						BaseMLDocument* item = new BaseMLDocument( checkPath( e->attribute( "document", "" ) ) );
						//item->setRecursion( getRecursion() );
						//item->setIndent( getIndent() );
						Stack callers;
						item->execute( os, action, 0, callers, attributes );
						delete item;
					}
				}
			}
		}//else elementnode
	}//for	
}





void
BaseMLDocument::renderContents( ostream& os, ElementNode* sub, ElementNode* parent, Stack& callers, Attributes& attributes ) {
  if(sub==0){
    cerr<<"sub is 0"<<endl;
    return;
  }
  for(XMLNode::iterator i=sub->begin();i!=sub->end();i++){
		if( (*i)->nodeType()==XMLNode::TextNode ) {
	    TextNode* e=(TextNode*) *i;
      if(e!=0){
        string data = e->data();
        data = attributes.apply( data );         
        os << removeWhiteSpace( data );
      }
		} else if( (*i)->nodeType()==XMLNode::PINode ) {
	    PINode* e = (PINode*) *i; //cast it
			string data = e->data();
			if( ( e->target() == "php" ) && ( PHPFilterDisabled == false ) ) {
			
				// create a temporary file to contain the php code
			
				string tempFileName;
				srand( time( NULL ) );
				for( int i = 0; i < 8; i++ )
					tempFileName += (char)( rand() % 26 + 97 );
				tempFileName += ".tmp";
				ofstream ofs( tempFileName.c_str() );
				ofs << "<?" << e->target() << " " << data << "?>";
				ofs.close();
				
				// execute php for the temporary file, and catch its output
				// we'll use a procbuffer for this
			
        /* we need to replace procbuf with popen here!!!!

				string command = "php -q " + tempFileName;
				procbuf pb( command.c_str(), ios::in );
				istream is( &pb );
				string s;
				while( getline( is, s ) ) os << s << endl;
				pb.close();
				
				// clean up the mess we made
				
				command = "rm -f " + tempFileName;
				system( command.c_str() );
			  */	
			} else
				os << "<?" << e->target() << " " << data << "?>";
		} else if( (*i)->nodeType()==XMLNode::ElementNode ) {
			ElementNode* e = (ElementNode*) *i;
			if( e!=0 ) { // the node was really an element
				string tagName = e->tagName();
				if( tagName == "space" ) {
					os << " ";
				} else
				if( tagName == "lt" ) {
					os << "&lt;";
				} else
				if( tagName == "gt" ) {
					os << "&gt;";
				} else
				if( tagName == "amp" ) {
					os << "&amp;";
				} else
				if( tagName == "quot" ) {
					os << "&quot;";
				} else
				if( tagName == "comment" ) {
					os << "<!--";
					execute( os, e, parent, callers, attributes );
					os << "-->";
				} else
				if( tagName == "indent" ) {
					//setIndent( getIndent() + 1 );
		      indent++;
					newLine( os );
					execute( os, e, parent, callers, attributes );
					//setIndent( getIndent() - 1 );
		      indent--;
					newLine( os );
				} else
				if( tagName == "cr" ) {
					newLine( os );
				} else
				if( tagName == "BML_credits" ) {
					os << "This file was generated with BaseML v" << VERSION;
				} else
				if( tagName == "BML_source" ) {
					os << "<a href=\"" << fileName << "\">" << fileName << "</a>";
				} else
				if( tagName == "BML_local_time" ) {
					time_t x;
					struct tm* t;
					::time(&x);
					t = localtime(&x);
					if( t->tm_hour < 10 )
						os << "0";
					os << t->tm_hour << ":";
					if( t->tm_min < 10 )
						os << "0";
					os << t->tm_min << ":";
					if( t->tm_sec < 10 )
						os << "0";
					os << t->tm_sec;
				} else
				if( tagName == "BML_gmt_time" ) {
					time_t x;
					struct tm* t;
					::time(&x);
					t = gmtime(&x);
					if( t->tm_hour < 10 )
						os << "0";
					os << t->tm_hour << ":";
					if( t->tm_min < 10 )
						os << "0";
					os << t->tm_min << ":";
					if( t->tm_sec < 10 )
						os << "0";
					os << t->tm_sec;
				} else
				if( tagName == "BML_gmt_date" ) {
					time_t x;
					struct tm* t;
					::time(&x);
					t = gmtime(&x);
					os << t->tm_mon+1 << "/" << t->tm_mday << "/" << t->tm_year+1900;
				} else
				if( tagName == "BML_local_date" ) {
					time_t x;
					struct tm* t;
					::time(&x);
					t = localtime(&x);
					os << t->tm_mon+1 << "/" << t->tm_mday << "/" << t->tm_year+1900;
				} else
				if(
					( tagName == "document" ) ||
					( tagName == "import" ) ||
					( tagName == "item" ) ||
					( tagName == "main" )
					) {
					cerr << "Warning: element \"" << tagName << "\" is a BaseML native element and cannot be used in user definitions" << endl;
				} else
				if( tagName == "caller" ) {
					if( callers.size() != 0 ) {
			      ElementNode* caller = callers.top();
						Stack newCallers( callers );
						newCallers.pop();
						renderContents( os, caller, NULL, newCallers, attributes );
					} else {
						cerr << "Warning: Cannot render contents of caller element in \"" << sub->tagName() << "\"." << endl;
					}
				} else
				if( tagName == "item" ) {
					cerr << "Warning: \"item\" elements can only reside in container elements" << endl;
				} else
				if( tagName == "define" ) {
					string name = e->attribute( "name", "" );
					if( name != "" ) {
						if( e->attribute( "scope", "" ) == "global" ) {
							BaseMLDocument::defines.global( name );
							execute( os, e, parent, callers, attributes );
						} else {
							BaseMLDocument::defines.push( name );
							execute( os, e, parent, callers, attributes );
							BaseMLDocument::defines.pop();
						}
					}
				} else
				if( tagName == "ifdef" ) {
					if( BaseMLDocument::defines.isDefined( e->attribute( "name", "" ) ) )
						execute( os, e, parent, callers, attributes );
				} else
				if( tagName == "ifndef" ) {
					if( !BaseMLDocument::defines.isDefined( e->attribute( "name", "" ) ) )
						execute( os, e, parent, callers, attributes );
				} else
				if( !filterHTML( os, e, tagName, sub, parent, callers, attributes ) ) {
					Stack newCallers( callers, e );
					Attributes newAttributes( attributes );
					newAttributes.addAttributes( e );
					execute( os, tagName, sub, newCallers, newAttributes );
				}
			}
		}
	}
}

/** Executes a sub program in the document */
void
BaseMLDocument::execute( ostream& os, ElementNode* sub, ElementNode* parent, Stack& callers, Attributes& attributes ) {
	//setRecursion( getRecursion() + 1 );
  recursion++;
	if( recursion > 200 ) {
		cerr << "Fatal Error: Recursion limit (200) reached" << endl;
		exit( 1 );
	}
	Attributes newAttributes( attributes );
	newAttributes.addAttributes( sub, false );
	if( sub->attribute("type", "function" ) == "container" ) {
		renderContainer( os, sub, parent, callers, newAttributes, callers.top()->attribute( "action", "main" ) );
	}	else
		renderContents( os, sub, parent, callers, newAttributes );
	//setRecursion( getRecursion() - 1 );
  recursion--;
}

/** Executes a sub program in the document */
void
BaseMLDocument::execute( ostream& os, const string& subName, ElementNode* parent, Stack& callers, Attributes& attributes ) {
	ElementNode* sub = findSub( subName );
	if(( sub == NULL )||(sub==0)) {
		cerr << "Warning: Cannot find definition of element \"" << subName << "\"." << endl;
	} else {
		execute( os, sub, parent, callers, attributes );
		//delete sub;
	}
}





/** Execute an element, and keep the original tags and attributes */
void
BaseMLDocument::executeTagged( ostream& os, ElementNode* sub, ElementNode* parent, Stack& callers, Attributes& attributes, bool evaluateContents ) {
	os << "<" << sub->tagName();
	for( unsigned int i = 0; i < sub->attributes.size(); i++ ) {
    string name = sub->attributes[i].name;
    string value = sub->attributes[i].value;
    value = attributes.apply( value );
    os << " " << name << "=\"" << value << "\"";
	}
	if( evaluateContents ) {
		os << ">";
		execute( os, sub, parent, callers, attributes );
		os << "</" << sub->tagName() << ">";
	} else {
		os << "/>";
	}
}




/** Strip whitespace from begin and end of string */
string
BaseMLDocument::removeWhiteSpace( const string& orig ){
  string result=orig;

  string::size_type a=0;
  while(  ((result[a]=='\t') ||
          (result[a]=='\n')) &&
          (a < result.length()) ) a++;

  string::size_type b=result.size()-1;
  while(  ((result[b]=='\t') ||
          (result[b]=='\n')) &&
          ( b>0 ) ) b--;

  return result.substr(a,b-a+1);
}


/** go to a new line in the output and indent */
void
BaseMLDocument::newLine( ostream& os ) {
	os << "\n";
	for( int i = 0; i < indent; i++ ) {
		os << "\t";
	}
}


/** Filters out an HTML tag, and if that's been done, returns TRUE. */
bool BaseMLDocument::filterHTML( ostream& os, ElementNode* e, const string& tagName, ElementNode* sub, ElementNode* parent, Stack& callers, Attributes& attributes ){
	if( !HTMLFilterDisabled ) {
   	if( // filter out html tags (empty elements)
   		( tagName == "br" ) ||
   		( tagName == "hr" ) ||
   		( tagName == "img" ) ||
   		( tagName == "meta" )
   	) {
   		executeTagged( os, e, parent, callers, attributes, false );
   		return true;
   	} else
   	if( // filter out html tags (elements with contents)
   		( tagName == "a" ) ||
   		( tagName == "abbr" ) ||
   		( tagName == "acronym" ) ||
   		( tagName == "address" ) ||
   		( tagName == "applet" ) ||
   		( tagName == "area" ) ||
   		( tagName == "b" ) ||
   		( tagName == "base" ) ||
   		( tagName == "basefont" ) ||
   		( tagName == "bdo" ) ||
   		( tagName == "big" ) ||
   		( tagName == "blockquote" ) ||
   		( tagName == "button" ) ||
   		( tagName == "body" ) ||
   		( tagName == "caption" ) ||
   		( tagName == "center" ) ||
   		( tagName == "cite" ) ||
   		( tagName == "code" ) ||
   		( tagName == "col" ) ||
   		( tagName == "colgroup" ) ||
   		( tagName == "dd" ) ||
   		( tagName == "del" ) ||
   		( tagName == "dfn" ) ||
   		( tagName == "dir" ) ||
   		( tagName == "div" ) ||
   		( tagName == "dl" ) ||
   		( tagName == "dt" ) ||
   		( tagName == "em" ) ||
   		( tagName == "fieldset" ) ||
   		( tagName == "font" ) ||
   		( tagName == "form" ) ||
   		( tagName == "frame" ) ||
   		( tagName == "frameset" ) ||
   		( tagName == "h1" ) ||
   		( tagName == "h2" ) ||
   		( tagName == "h3" ) ||
   		( tagName == "h4" ) ||
   		( tagName == "h5" ) ||
   		( tagName == "h6" ) ||
   		( tagName == "head" ) ||
   		( tagName == "html" ) ||
   		( tagName == "i" ) ||
   		( tagName == "iframe" ) ||
   		( tagName == "input" ) ||
   		( tagName == "ins" ) ||
   		( tagName == "isindex" ) ||
   		( tagName == "kbd" ) ||
   		( tagName == "label" ) ||
   		( tagName == "legend" ) ||
   		( tagName == "li" ) ||
   		( tagName == "link" ) ||
   		( tagName == "ol" ) ||
   		( tagName == "map" ) ||
   		( tagName == "menu" ) ||
   		( tagName == "noframes" ) ||
   		( tagName == "noscript" ) ||
   		( tagName == "object" ) ||
   		( tagName == "option" ) ||
   		( tagName == "optgroup" ) ||
   		( tagName == "p" ) ||
   		( tagName == "param" ) ||
   		( tagName == "pre" ) ||
   		( tagName == "q" ) ||
   		( tagName == "s" ) ||
   		( tagName == "samp" ) ||
   		( tagName == "script" ) ||
   		( tagName == "select" ) ||
   		( tagName == "small" ) ||
   		( tagName == "span" ) ||
   		( tagName == "strike" ) ||
   		( tagName == "strong" ) ||
   		( tagName == "style" ) ||
   		( tagName == "sub" ) ||
   		( tagName == "sup" ) ||
   		( tagName == "table" ) ||
   		( tagName == "tbody" ) ||
   		( tagName == "td" ) ||
   		( tagName == "textarea" ) ||
   		( tagName == "tfoot" ) ||
   		( tagName == "th" ) ||
   		( tagName == "thead" ) ||
   		( tagName == "title" ) ||
   		( tagName == "tr" ) ||
   		( tagName == "tt" ) ||
   		( tagName == "u" ) ||
   		( tagName == "ul" ) ||
   		( tagName == "var" )
   	) {
   		executeTagged( os, e, parent, callers, attributes, true );
   		return true;
   	}
  }
	return false;
}



string BaseMLDocument::checkPath( const string& fileName ){
  if( fileName[0] == '~' ) {
		string result = fileName;
	  result=getenv( "HOME" ) + result.substr( 1, result.length()-1 );
		return result;
	} else
		if( fileName[0] != '/' )
			return filePath + fileName;
		else
			return fileName;
}

