/***************************************************************************
                          basemldocument.h  -  description
                             -------------------
    begin                : Sat Dec 8 2001
    copyright            : (C) 2001-2002 by Anthony Liekens
    email                : mooby@nerdhero.org
    Modified by          : Walter Schreppers
    email                : wschrep@uia.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef BASEMLDOCUMENT_H
#define BASEMLDOCUMENT_H

#include "stack.h"
#include "attributes.h"
#include "defines.h"

#include "XMLParser.h"


/**
  *@author Anthony Liekens,Walter Schreppers
  */

	static int recursion;
	static int indent;

class BaseMLDocument {

protected:
	bool HTMLFilterDisabled, PHPFilterDisabled;
  string fileName;
  string filePath;
  
  XMLNode* tree, * importDoc;
  XMLParser xml;
  
	BaseMLDocument* parentDocument; // we will use this in case we need something from an extended document

public:

	static Defines defines;

	BaseMLDocument( const string& fileName, const string& childName = "" );
	~BaseMLDocument();
	
	/** Finds a sub program in the document */
	ElementNode* findSub( const string &subName );
	/** check if document has a parent, and if not loaded, load it */
	bool hasParent();
  /** Executes a sub program in the document */
  void execute( ostream& os, const string& subName, ElementNode* parent, Stack& callers, Attributes& attributes );
  void execute( ostream& os, ElementNode* sub, ElementNode* parent, Stack& callers, Attributes &attributes );
  void executeTagged( ostream& os, ElementNode* sub, ElementNode* parent, Stack& callers, Attributes& attributes, bool evaluateContents = true );
  /** Render the contents of a function */
  void renderContents( ostream& os, ElementNode* sub, ElementNode* parent,  Stack& callers, Attributes& attributes );
  /** Render the contents of a container */
  void renderContainer( ostream& os, ElementNode* sub, ElementNode* parent,  Stack& callers, Attributes& attributes, const string& action );
  /** Filters out an HTML tag, and if that's been done, returns TRUE. */
  bool filterHTML( ostream& os, ElementNode* e, const string& tagName, ElementNode* sub, ElementNode* parent, Stack& callers, Attributes& attributes );

  /** Return the node type of a node */
  //string nodeType( const QDomNode& );
  /** Strip whitespace */
  string removeWhiteSpace( const string& );

  void newLine( ostream& os );

  int getRecursion() { return recursion; }
  int getIndent() { return indent; }
  /** No descriptions */
  void importDocument( const string& fileName );

  string checkPath(const string& fileName );

};

#endif
