#include "options.h"
#include <iostream>
using namespace std;

Option::Option( string name, string comments, bool hasArgument, bool isRequired ) {

	setName( name );
	setComments( comments );
	setHasArgument( hasArgument );
	setIsRequired( isRequired );

	setIsParsed( false );
	setArgument( "" );

}

void
Options::setExtras( bool extras, string extraName, int extraLimit, int extraRequired ) {
	this->extras = extras;
	this->extraName = extraName;
	this->extraLimit = extraLimit;
	this->extraRequired = extraRequired;
}

vector< string >
Options::parse( int argc, char** argv ) {

	// create a vector of strings to contain the return value

	vector< string> filtered;
	
	// parse the argument values
	
	for( int count = 1; count < argc; count++ ) {
	
		string parameter( argv[ count ] );
		
		if( parameter[0] != '-' ) { 
		
			if( extras ) {
		
				// this is not a parameter, but something that might prove useful for the user
				
				filtered.push_back( parameter );
				
				if( extraLimit != -1 )
				if( filtered.size() > extraLimit ) {
					cerr << "Too many " << extraName << " parameters" << endl;
					printUsage( cerr, argv[0] );
					exit( 1 );
				}
			
			} else {
			
				cerr << "Cannot process parameter " << parameter << endl;
				printUsage( cerr, argv[0] );
				exit(1);

			}
			
		} else if( parameter.length() == 1 ) {
		
			// we don't know what to do with parameter "-"
			
			
		} else if( parameter.length() == 2 ) {
		
			// we have something of the form "-x", see if we got an option starting with 'x'
			
			iterator i = begin();
			bool found = false;
			
			do {
			
				if( i->getName()[0] == parameter[1] ) {
				
					i->setIsParsed( true );
					found = true;
					
					if( i->getHasArgument() ) { 
					
						// this parameter needs an extra argument
						
						if( count != argc - 1 ) {
							count++;
							i->setArgument( string( argv[ count ] ) );
						} else {
							cerr << "Parameter " << parameter << " needs an argument" << endl;
							printUsage( cerr, argv[ 0 ] );
							exit( 1 );
						}
					
					} 
					
				}
				
				if( i != end() )
					i++;
				
			} while( !found && ( i != end() ) );
			
			if( !found ) {
				cerr << "Cannot process parameter " << parameter << endl;
				printUsage( cerr, argv[ 0 ] );
				exit( 1 );
			}
			
		} else if( (parameter[1] == '-') && ( parameter.length() > 2 ) ) {
			// we have something of the form "--foo", see if we got an option 'foo'
			
			iterator i = begin();
			bool found = false;
			
			do {
			
				string parameterCutted = parameter.substr( 2, parameter.length() - 2 );
			
				if( i->getName() == parameterCutted ) {
				
					i->setIsParsed( true );
					found = true;
					
					if( i->getHasArgument() ) { 
					
						// this parameter needs an extra argument
						
						if( count != argc - 1 ) {
							count++;
							i->setArgument( string( argv[ count ] ) );
						} else {
							cerr << "Parameter " << parameter << " requires an argument" << endl;
							printUsage( cerr, argv[ 0 ] );
							exit( 1 );
						}
					
					} 
					
				}
				
				if( i != end() )
					i++;
				
			} while( !found && ( i != end() ) );
			
			if( !found ) {
				cerr << "Cannot process parameter " << parameter << endl;
				printUsage( cerr, argv[ 0 ] );
				exit( 1 );
			}
			
		} else {
			cerr << "Cannot process parameter " << parameter << endl;
			printUsage( cerr, argv[ 0 ] );
			exit( 1 );
		}
	}
	
	// see if the user missed a required argument
	
	for( iterator i = begin(); i != end(); i++ ) {
		if( ( i->getIsRequired() ) && ( !i->getIsParsed() ) ) {
			cerr << "Parameter --" << i->getName() << " was required but is missing" << endl;
			printUsage( cerr, argv[ 0 ] );
			exit( 1 );
		}
	}
	
	// see if the user missed an extra argument
	
	if( extras ) {

		if( filtered.size() < extraRequired ) {
		
			cerr << "Not enough " << extraName << " parameters" << endl;
			printUsage( cerr, argv[ 0 ] );
			exit( 1 );
		
		}

	}
	
	return filtered;
}

void
Options::printUsage( ostream& os, string programName ) {
	os << "Usage: " << programName << " ";
	for( iterator i = begin(); i != end(); i++ ) {
		if( !i->getIsRequired() ) {
			os << "[";
		}
		os << "--" << i->getName();
		if( i->getHasArgument() ) {
			os << " param";
		}
		if( !i->getIsRequired() ) {
			os << "]";
		}
		os << " ";
	}
	if( extras ) {
		if( extraRequired > 0 )
		for( int i = 0; i < extraRequired; i++ ) {
			os << extraName << " ";
		}
		if( extraLimit == -1 ) {
			os << "{ " << extraName << " ... } ";
		} else {
			for( int i = 0; i < extraLimit - extraRequired; i++ ) {
				os << "{ " << extraName << " ";
			}				
			for( int i = 0; i < extraLimit - extraRequired; i++ ) {
				os << "} ";
			}				
			
		}
	}
	os << endl;
	os << "Parameters:" << endl;
	for( iterator i = begin(); i != end(); i++ ) {
		os << "\t--" << i->getName() << ", -" << i->getName()[0];
		os << "\t" << i->getComments() << endl;
	}
}
