#ifndef OPTIONS_H
#define OPTIONS_H

#include <vector>
#include <string>
#include <iostream>

using namespace std;

class Option {

protected:

	string name, comments, argument;
	bool isRequired, hasArgument, isParsed;

public:

	enum requiredOptions { REQUIRED = true, NOTREQUIRED = false };
	enum argumentOptions { ARGUMENT = true, NOARGUMENT = false };

	Option() {}
	Option( string name, string comments, bool hasArgument = false, bool isRequired = false );
	
	void setName( string name ) { this->name = name; }
	void setComments( string comments ) { this->comments = comments; }
	void setArgument( string argument ) { this->argument = argument; }
	void setIsRequired( bool isRequired ) { this->isRequired = isRequired; }
	void setHasArgument( bool hasArgument ) { this->hasArgument = hasArgument; }
	void setIsParsed( bool isParsed ) { this->isParsed = isParsed; }
	
	string getName() { return name; }
	string getComments() { return comments; }
	string getArgument() { return argument; }
	bool getIsRequired() { return isRequired; }
	bool getHasArgument() { return hasArgument; }
	bool getIsParsed() { return isParsed; }
	
};

class Options : public vector< Option > {

bool extras;
string extraName;
int extraLimit, extraRequired;

public:	

	Options() {
		extras = false;
	}
	
	/**
		set if any extra options are possible
		this can e.g.be used for files or extra parameters
	*/
	
	void setExtras( bool extras, string extraName = "file", int extraLimit = 0, int extraRequired = 0 );
	
	/**
		parse the command line options
		if anything non-conformant with the options is found, it is put in the
		returned vector, which allows the users to use these as e.g. lists of 
		files for input, etc.
	*/
	
	vector< string > parse( int, char** );
	
	/**
		Print usage message to ostream
	*/
	
	void printUsage( ostream& os, string programName );
	
	/**
		Check whether a parameter was included
	*/
	
	bool 
	optionGiven( string parameter ) {
		for( iterator i = begin(); i != end(); i++ ) {
			if( i->getName() == parameter )
				return i->getIsParsed();
		}
		return false;
	}
	
	/**
		Return the argument given
	*/
	
	string
	getOption( string parameter ) {
		for( iterator i = begin(); i != end(); i++ ) {
			if( i->getName() == parameter )
				return i->getArgument();
		}
		return "";
	}
	
};

#endif
