#include "xmlnode.h"


XMLNode::XMLNode(){
  clear();
  parent=0;
}

XMLNode::XMLNode(XMLNode* p){
  clear();
  setParent(p);
}


void XMLNode::setParent(XMLNode* p){
  parent=p;
}


XMLNode::~XMLNode(){
  destroy(this);
}

//
//recursively walk down tree and delete every node bottom up
//
void XMLNode::destroy(XMLNode* node){
  if( (node!=0) && (node->size()>0) ) {
    for( XMLNode::iterator i=node->begin(); i!= node->end(); i++ ){
      
      destroy( *i );
      //cout<<"deleting:"<<(*i)->getName()<<endl;
      
      (*i)->clear(); //free children
      //delete ( *i ); //free mem
    }
  }
}


void XMLNode::setName(const string& s){
  name=s;
}

string XMLNode::getName() const{
  return name;
}


void XMLNode::show(int indent){
  string s="";
  for(int i=0;i<indent;i++) s+="  ";
  s+=getName();
  cout<<s<<endl;
}

//
//recursively walk through tree and show node names with indentation
//
void XMLNode::showTree(XMLNode* node, int indent) const{
  indent++;
  if( (node!=0) && (node->size()>0) ){
    for( XMLNode::const_iterator i=node->begin(); i!= node->end(); i++ ){
      (*i)->show( indent );
      showTree( *i , indent );
    }
  }
}



void XMLNode::appendChild(XMLNode* node){
  node->setParent(this);
  push_back(node);
}

void XMLNode::appendSibling(XMLNode* node){
  node->parent=parent;
  if(parent!=0){
    parent->push_back(node);
  }
}

XMLNode::iterator XMLNode::lookup(){
  if(parent!=0){
    XMLNode::iterator i=parent->begin();
    while((*i!=this)&&(i!=parent->end())){
      i++;
    }
    return i;
  }
  return end();
}




//returns the nextSibling
XMLNode* XMLNode::nextSibling(){
  if(parent!=0){
    XMLNode::iterator i=lookup();
    if(i!=parent->end()){
      i++;
      return *i;
    }
  }
  return 0;
}



XMLNode* XMLNode::prevSibling(){
  if(parent!=0){
    XMLNode::iterator i=lookup();
    if(i!=parent->end()){
      i--;
      return *i;
    }
  }
  return 0;
}



//returns first child of a node
XMLNode* XMLNode::firstChild(){
  XMLNode::const_iterator child=this->begin();
  return *child;
}



//returns last child of a node
XMLNode* XMLNode::lastChild(){
  XMLNode::const_iterator child=this->end();
  child--;
  return *child;
}

